@@ -57,14 +57,15 @@ enum {
 #define DEFAULT_MAX_HISTORY_SAVED 10  /* This seems to make more sense then 60*/
 
 struct _GnomeEntryPrivate {
-	gchar     *history_id;
+	gchar       *history_id;
 
-	GList     *items;
+	GList       *items;
 
-	guint16    max_saved;
-	guint32    changed : 1;
-
-	guint      gconf_notify_id;
+	guint16      max_saved;
+	guint        changed : 1;
+	guint        saving_history : 1;
+	GConfClient *gconf_client;
+	guint        gconf_notify_id;
 };
 
 
@@ -77,6 +78,7 @@ struct item {
 static void gnome_entry_class_init (GnomeEntryClass *class);
 static void gnome_entry_init       (GnomeEntry      *gentry);
 static void gnome_entry_finalize   (GObject         *object);
+static void gnome_entry_destroy    (GtkObject       *object);
 
 static void gnome_entry_get_property (GObject        *object,
 				      guint           param_id,
@@ -101,26 +103,27 @@ gnome_entry_get_type (void)
 	static GType object_type = 0;
 
 	if (object_type == 0) {
-		GtkType type_of_parent;
-		static const GtkTypeInfo object_info = {
-			"GnomeEntry",
-			sizeof (GnomeEntry),
+		static const GTypeInfo object_info = {
 			sizeof (GnomeEntryClass),
-			(GtkClassInitFunc) gnome_entry_class_init,
-			(GtkObjectInitFunc) gnome_entry_init,
-			/* reserved_1 */ NULL,
-			/* reserved_2 */ NULL,
-			(GtkClassInitFunc) NULL
+			(GBaseInitFunc) NULL,
+			(GBaseFinalizeFunc) NULL,
+			(GClassInitFunc) gnome_entry_class_init,
+			(GClassFinalizeFunc) NULL,
+			NULL,			/* class_data */
+			sizeof (GnomeEntry),
+			0,			/* n_preallocs */
+			(GInstanceInitFunc) gnome_entry_init,
+			NULL			/* value_table */
 		};
+
 		static const GInterfaceInfo editable_info =
 		{
 			(GInterfaceInitFunc) gnome_entry_editable_init,	 /* interface_init */
 			NULL,			                         /* interface_finalize */
 			NULL			                         /* interface_data */
 		};
-		type_of_parent = GTK_TYPE_COMBO;
-		object_type = gtk_type_unique (type_of_parent, &object_info);
-		parent_class = gtk_type_class (type_of_parent);
+
+		object_type = g_type_register_static (GTK_TYPE_COMBO, "GnomeEntry", &object_info, 0);
 
 		g_type_add_interface_static (object_type,
 					     GTK_TYPE_EDITABLE,
@@ -130,47 +133,60 @@ gnome_entry_get_type (void)
 }
 
 enum {
-	CHANGED_SIGNAL,
 	ACTIVATE_SIGNAL,
 	LAST_SIGNAL
 };
 static int gnome_entry_signals[LAST_SIGNAL] = {0};
 
+static gboolean
+gnome_entry_mnemonic_activate (GtkWidget *widget,
+			       gboolean   group_cycling)
+{
+	gboolean handled;
+	GnomeEntry *entry;
+	
+	entry = GNOME_ENTRY (widget);
+
+	group_cycling = group_cycling != FALSE;
+
+	if (!GTK_WIDGET_IS_SENSITIVE (GTK_COMBO (entry)->entry))
+		handled = TRUE;
+	else
+		g_signal_emit_by_name (GTK_COMBO (entry)->entry, "mnemonic_activate", group_cycling, &handled);
+
+	return handled;
+}
+
 static void
 gnome_entry_class_init (GnomeEntryClass *class)
 {
 	GtkObjectClass *object_class;
+	GtkWidgetClass *widget_class;
 	GObjectClass *gobject_class;
 
+	parent_class = g_type_class_peek_parent (class);
+
 	object_class = (GtkObjectClass *) class;
 	gobject_class = (GObjectClass *) class;
-
-	gnome_entry_signals[CHANGED_SIGNAL] =
-		gtk_signal_new("changed",
-			       GTK_RUN_LAST,
-			       GTK_CLASS_TYPE (object_class),
-			       GTK_SIGNAL_OFFSET(GnomeEntryClass,
-			       			 changed),
-			       gtk_signal_default_marshaller,
-			       GTK_TYPE_NONE,
-			       0);
+	widget_class = (GtkWidgetClass *) class;
+	
 	gnome_entry_signals[ACTIVATE_SIGNAL] =
-		gtk_signal_new("activate",
-			       GTK_RUN_LAST,
-			       GTK_CLASS_TYPE (object_class),
-			       GTK_SIGNAL_OFFSET(GnomeEntryClass,
-			       			 activate),
-			       gtk_signal_default_marshaller,
-			       GTK_TYPE_NONE,
-			       0);
+		g_signal_new("activate",
+			     G_TYPE_FROM_CLASS (gobject_class),
+			     G_SIGNAL_RUN_LAST,
+			     G_STRUCT_OFFSET(GnomeEntryClass, activate),
+			     NULL, NULL,
+			     g_cclosure_marshal_VOID__VOID,
+			     G_TYPE_NONE, 0);
 
-	class->changed = NULL;
 	class->activate = NULL;
 
+	object_class->destroy = gnome_entry_destroy;
 	gobject_class->finalize = gnome_entry_finalize;
 	gobject_class->set_property = gnome_entry_set_property;
 	gobject_class->get_property = gnome_entry_get_property;
-
+	widget_class->mnemonic_activate = gnome_entry_mnemonic_activate;
+	
 	g_object_class_install_property (gobject_class,
 					 PROP_HISTORY_ID,
 					 g_param_spec_string ("history_id",
@@ -206,7 +222,7 @@ gnome_entry_get_property (GObject        *object,
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
 		break;
 	}
-	
+
 }
 
 static void
@@ -222,12 +238,12 @@ gnome_entry_set_property (GObject        *object,
 		gnome_entry_set_history_id (entry, g_value_get_string (value));
 		gnome_entry_load_history (entry);
 		break;
-		
+
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
 		break;
 	}
-	
+
 }
 
 static void
@@ -238,7 +254,7 @@ entry_changed (GtkWidget *widget, gpointer data)
 	gentry = data;
 	gentry->_priv->changed = TRUE;
 
-	gtk_signal_emit (GTK_OBJECT (gentry), gnome_entry_signals[CHANGED_SIGNAL]);
+	g_signal_emit_by_name (gentry, "changed");
 }
 
 static void
@@ -257,8 +273,7 @@ entry_activated (GtkWidget *widget, gpointer data)
 		gnome_entry_prepend_history (gentry, TRUE, gtk_entry_get_text (GTK_ENTRY (widget)));
 	}
 
-	gtk_signal_emit (GTK_OBJECT (gentry), gnome_entry_signals[ACTIVATE_SIGNAL]);
-	gnome_entry_save_history (gentry);
+	g_signal_emit (gentry, gnome_entry_signals[ACTIVATE_SIGNAL], 0);
 }
 
 static void
@@ -271,12 +286,12 @@ gnome_entry_init (GnomeEntry *gentry)
 	gentry->_priv->items        = NULL;
 	gentry->_priv->max_saved    = DEFAULT_MAX_HISTORY_SAVED;
 
-	gtk_signal_connect (GTK_OBJECT (gnome_entry_gtk_entry (gentry)), "changed",
-			    (GtkSignalFunc) entry_changed,
-			    gentry);
-	gtk_signal_connect (GTK_OBJECT (gnome_entry_gtk_entry (gentry)), "activate",
-			    (GtkSignalFunc) entry_activated,
-			    gentry);
+	g_signal_connect (gnome_entry_gtk_entry (gentry), "changed",
+			  G_CALLBACK (entry_changed),
+			  gentry);
+	g_signal_connect (gnome_entry_gtk_entry (gentry), "activate",
+			  G_CALLBACK (entry_activated),
+			  gentry);
 	gtk_combo_disable_activate (GTK_COMBO (gentry));
         gtk_combo_set_case_sensitive (GTK_COMBO (gentry), TRUE);
 }
@@ -324,34 +339,59 @@ free_items (GnomeEntry *gentry)
 }
 
 static void
-gnome_entry_finalize (GObject *object)
+gnome_entry_destroy (GtkObject *object)
 {
 	GnomeEntry *gentry;
-	GConfClient *client;
-	gchar *key;
-	
+
+	/* Note: destroy can run multiple times */
+
 	g_return_if_fail (object != NULL);
 	g_return_if_fail (GNOME_IS_ENTRY (object));
 
 	gentry = GNOME_ENTRY (object);
 
-	gnomeui_gconf_lazy_init ();
-	client = gconf_client_get_default ();
-	key = build_gconf_key (gentry);
-	gconf_client_remove_dir (client, key, NULL);
-	g_free (key);
-	
-	if (gentry->_priv->gconf_notify_id != 0) {
-		gconf_client_notify_remove (client, gentry->_priv->gconf_notify_id);
+	if (gentry->_priv->gconf_client != NULL) {
+		gchar *key;
+
+		if (gentry->_priv->gconf_notify_id != 0) {
+			gconf_client_notify_remove
+				(gentry->_priv->gconf_client,
+				 gentry->_priv->gconf_notify_id);
+			gentry->_priv->gconf_notify_id = 0;
+		}
+
+		key = build_gconf_key (gentry);
+		gconf_client_remove_dir (gentry->_priv->gconf_client,
+					 key, NULL);
+		g_free (key);
+
+		g_object_unref (G_OBJECT (gentry->_priv->gconf_client));
+
+		gentry->_priv->gconf_client = NULL;
 	}
 
-	g_free(gentry->_priv);
+	GNOME_CALL_PARENT (GTK_OBJECT_CLASS, destroy, (object));
+}
+
+static void
+gnome_entry_finalize (GObject *object)
+{
+	GnomeEntry *gentry;
+
+	g_return_if_fail (object != NULL);
+	g_return_if_fail (GNOME_IS_ENTRY (object));
+
+	gentry = GNOME_ENTRY (object);
+
+	g_free (gentry->_priv->history_id);
+	gentry->_priv->history_id = NULL;
+
+	free_items (gentry);
+
+	g_free (gentry->_priv);
 	gentry->_priv = NULL;
 
-	g_object_unref (G_OBJECT (client));
-	
-	if (G_OBJECT_CLASS (parent_class)->finalize)
-		(* G_OBJECT_CLASS (parent_class)->finalize) (object);
+	GNOME_CALL_PARENT (G_OBJECT_CLASS, finalize, (object));
 }
 
 /**
@@ -381,42 +421,66 @@ gnome_entry_history_changed (GConfClient* client,
 
 	gentry = GNOME_ENTRY (user_data);
 
+	/* If we're getting a notification from saving our own
+	 * history, don't reload it.
+	 */
+	if (gentry->_priv->saving_history) {
+		gentry->_priv->saving_history = FALSE;
+
+		return;
+	}
+
 	gnome_entry_load_history (gentry);
 }
 
 /* FIXME: Make this static */
+
+/**
+ * gnome_entry_set_history_id
+ * @gentry: Pointer to GnomeEntry object.
+ * @history_id: the text id under which history data is stored
+ *
+ * Description: Set the id of the history list. This function cannot be
+ * used to change and already existing id.
+ */
 void
 gnome_entry_set_history_id (GnomeEntry *gentry, const gchar *history_id)
 {
 	gchar *key;
-	GConfClient *client;
-	
+
 	g_return_if_fail (gentry != NULL);
 	g_return_if_fail (GNOME_IS_ENTRY (gentry));
-	g_return_if_fail (gentry->_priv->history_id == NULL);
 
+	if (gentry->_priv->history_id != NULL) {
+		g_warning ("The program is trying to change an existing "
+			   "GnomeEntry history id. This operation is "
+			   "not allowed.");
+
+		return;
+	}
+				
 	if (history_id == NULL)
 		return;
 
-	gentry->_priv->history_id = g_strdup (history_id); /* this handles NULL correctly */
+	gentry->_priv->history_id = g_strdup (history_id);
 
 	/* Register with gconf */
 	key = build_gconf_key (gentry);
-	gnomeui_gconf_lazy_init ();
-	client = gconf_client_get_default ();
 
-	gconf_client_add_dir (client,
+	gnomeui_gconf_lazy_init ();
+	gentry->_priv->gconf_client = gconf_client_get_default ();
+
+	gconf_client_add_dir (gentry->_priv->gconf_client,
 			      key,
 			      GCONF_CLIENT_PRELOAD_NONE,
 			      NULL);
-	
-	gentry->_priv->gconf_notify_id = gconf_client_notify_add (client,
+
+	gentry->_priv->gconf_notify_id = gconf_client_notify_add (gentry->_priv->gconf_client,
 								  key,
 								  gnome_entry_history_changed,
 								  gentry,
 								  NULL, NULL);
 
-	g_object_unref (G_OBJECT (client));
 	g_free (key);
 }
 
@@ -444,11 +508,9 @@ gnome_entry_get_history_id (GnomeEntry *gentry)
  * @max_saved: Maximum number of history items to save
  *
  * Description: Set internal limit on number of history items saved
- * to the config file, when #gnome_entry_save_history() is called.
+ * to the config file.
  * Zero is an acceptable value for @max_saved, but the same thing is
  * accomplished by setting the history id of @gentry to %NULL.
- *
- * Returns:
  */
 void
 gnome_entry_set_max_saved (GnomeEntry *gentry, guint max_saved)
@@ -464,7 +526,7 @@ gnome_entry_set_max_saved (GnomeEntry *gentry, guint max_saved)
  * @gentry: Pointer to GnomeEntry object.
  *
  * Description: Get internal limit on number of history items saved
- * to the config file, when #gnome_entry_save_history() is called.
+ * to the config file
  * See #gnome_entry_set_max_saved().
  *
  * Returns: An unsigned integer
@@ -481,52 +543,112 @@ gnome_entry_get_max_saved (GnomeEntry *gentry)
 static char *
 build_gconf_key (GnomeEntry *gentry)
 {
-	return g_strconcat ("/apps/",
-			    gnome_program_get_app_id (gnome_program_get()),
-			    "/history-",
-			    gentry->_priv->history_id,
-			    NULL);
+	gchar *retval;
+	gchar *app_id;
+
+	app_id = gconf_escape_key (gnome_program_get_app_id (gnome_program_get()), -1);
+
+	retval = g_strconcat ("/apps/gnome-settings/",
+			      gnome_program_get_app_id (gnome_program_get()),
+			      "/history-",
+			      gentry->_priv->history_id,
+			      NULL);
+
+	g_free (app_id);
+
+	return retval;
+}
+
+static void
+set_combo_items (GnomeEntry *gentry)
+{
+	GnomeEntryPrivate *priv;
+	GList *l;
+	char *text;
+
+	priv = gentry->_priv;
+
+	/* Save the contents of the entry so that setting the new list of
+	 * strings will not overwrite it.
+	 */
+
+	text = gtk_editable_get_chars (GTK_EDITABLE (GTK_COMBO (gentry)->entry), 0, -1);
+
+	/* Build the list of strings out of our items */
+
+	if (priv->items) {
+		GList *strings = NULL;
+
+		for (l = priv->items; l; l = l->next) {
+			struct item *item = l->data;
+
+			strings = g_list_prepend (strings, item->text);
+		}
+
+		strings = g_list_reverse (strings);
+	
+		gtk_combo_set_popdown_strings (GTK_COMBO (gentry), strings);
+
+		g_list_free (strings);
+	}
+
+	/* Restore the text in the entry and clear our changed flag. */
+
+	gtk_entry_set_text (GTK_ENTRY (GTK_COMBO (gentry)->entry), text);
+	g_free (text);
+
+	priv->changed = FALSE;
 }
 
 static void
 gnome_entry_add_history (GnomeEntry *gentry, gboolean save,
 			 const gchar *text, gboolean append)
 {
+	GnomeEntryPrivate *priv;
 	struct item *item;
-	GList *gitem;
-	GtkWidget *li;
-	GtkWidget *entry;
-	gchar *tmp;
+	GList *list;
+	gboolean changed;
 
 	g_return_if_fail (gentry != NULL);
 	g_return_if_fail (GNOME_IS_ENTRY (gentry));
 	g_return_if_fail (text != NULL); /* FIXME: should we just return without warning? */
 
-	entry = gnome_entry_gtk_entry (gentry);
-	tmp = g_strdup (gtk_entry_get_text (GTK_ENTRY (entry)));
-	
+	priv = gentry->_priv;
+
 	item = g_new (struct item, 1);
 	item->save = save;
 	item->text = g_strdup (text);
 
-	gentry->_priv->items = g_list_prepend (gentry->_priv->items, item);
+	/* Remove duplicates */
+	changed = FALSE;
+	list = priv->items;
+	
+	while (list != NULL) {
+		struct item * data = list->data;
+			
+		if (strcmp (data->text, text) == 0) {
+			free_item (data, NULL);
+			priv->items = g_list_delete_link (priv->items, list);
+			
+			changed = TRUE;
+			break;
+		}
 
-	li = gtk_list_item_new_with_label (text);
-	gtk_widget_show (li);
-
-	gitem = g_list_append (NULL, li);
+		list = g_list_next (list);
+	}
 
+	/* Really add the history item */
 	if (append)
-	  gtk_list_prepend_items (GTK_LIST (GTK_COMBO (gentry)->list), gitem);
+		priv->items = g_list_append (priv->items, item);
 	else
-	  gtk_list_append_items (GTK_LIST (GTK_COMBO (gentry)->list), gitem);
+		priv->items = g_list_prepend (priv->items, item);
 
-	gtk_entry_set_text (GTK_ENTRY (entry), tmp);
-	g_free (tmp);
+	/* Update combo list */
+	set_combo_items (gentry);
 
-	/* gtk_entry_set_text runs our 'entry_changed' routine, so we have
-	   to undo the effect */
-	gentry->_priv->changed = FALSE;
+	/* Save the history if needed */
+	if (changed || save)
+		gnome_entry_save_history (gentry);
 }
 
 
@@ -541,14 +663,18 @@ gnome_entry_add_history (GnomeEntry *gentry, gboolean save,
  * the history list inside @gentry.  If @save is %TRUE, the history
  * item will be saved in the config file (assuming that @gentry's
  * history id is not %NULL).
- *
- * Returns:
+ * Duplicates are automatically removed from the history list.
+ * The history list is automatically saved if needed.
  */
 void
 gnome_entry_prepend_history (GnomeEntry *gentry, gboolean save,
 			     const gchar *text)
 {
-  gnome_entry_add_history (gentry, save, text, FALSE);
+	g_return_if_fail (gentry != NULL);
+	g_return_if_fail (GNOME_IS_ENTRY (gentry));
+	g_return_if_fail (text != NULL); /* FIXME: should we just return without warning? */
+	
+	gnome_entry_add_history (gentry, save, text, FALSE);
 }
 
 
@@ -562,83 +688,44 @@ gnome_entry_prepend_history (GnomeEntry *gentry, gboolean save,
  * of the history list inside @gentry.  If @save is %TRUE, the
  * history item will be saved in the config file (assuming that
  * @gentry's history id is not %NULL).
- *
- * Returns:
+ * Duplicates are automatically removed from the history list.
+ * The history list is automatically saved if needed.
  */
 void
 gnome_entry_append_history (GnomeEntry *gentry, gboolean save,
 			    const gchar *text)
 {
+	g_return_if_fail (gentry != NULL);
+	g_return_if_fail (GNOME_IS_ENTRY (gentry));
+	g_return_if_fail (text != NULL); /* FIXME: should we just return without warning? */
+
 	gnome_entry_add_history (gentry, save, text, TRUE);
 }
 
 
-
-static void
-set_combo_items (GnomeEntry *gentry)
-{
-	GtkList *gtklist;
-	GList *items;
-	GList *gitems;
-	struct item *item;
-	GtkWidget *li;
-	GtkWidget *entry;
-	gchar *tmp;
-
-	gtklist = GTK_LIST (GTK_COMBO (gentry)->list);
-
-	/* We save the contents of the entry because when we insert
-	 * items on the combo list, the contents of the entry will get
-	 * changed.
-	 */
-
-	entry = gnome_entry_gtk_entry (gentry);
-	tmp = g_strdup (gtk_entry_get_text (GTK_ENTRY (entry)));
-	
-	gtk_list_clear_items (gtklist, 0, -1); /* erase everything */
-
-	gitems = NULL;
-
-	for (items = gentry->_priv->items; items; items = items->next) {
-		item = items->data;
-
-		li = gtk_list_item_new_with_label (item->text);
-		gtk_widget_show (li);
-		
-		gitems = g_list_append (gitems, li);
-	}
-
-	gtk_list_append_items (gtklist, gitems); /* this handles NULL correctly */
-
-	gtk_entry_set_text (GTK_ENTRY (entry), tmp);
-	g_free (tmp);
-
-	gentry->_priv->changed = FALSE;
-}
-
-
 static void
 gnome_entry_load_history (GnomeEntry *gentry)
 {
 	struct item *item;
 	gchar *key;
 	GSList *gconf_items, *items;
-	GConfClient *client;
-	
+
 	g_return_if_fail (gentry != NULL);
 	g_return_if_fail (GNOME_IS_ENTRY (gentry));
 
-	if (gnome_program_get_app_id (gnome_program_get()) == NULL ||gentry->_priv->history_id == NULL)
+	if (gnome_program_get_app_id (gnome_program_get()) == NULL ||
+	    gentry->_priv->history_id == NULL)
 		return;
 
+	g_return_if_fail (gentry->_priv->gconf_client != NULL);
+
 	free_items (gentry);
 
 	key = build_gconf_key (gentry);
 
-	gnomeui_gconf_lazy_init ();
-	client = gconf_client_get_default ();
-	
-	gconf_items = gconf_client_get_list (client, key, GCONF_VALUE_STRING, NULL);
+	gconf_items = gconf_client_get_list (gentry->_priv->gconf_client,
+					     key, GCONF_VALUE_STRING, NULL);
+	g_free (key);
 
 	for (items = gconf_items; items; items = items->next) {
 
@@ -652,17 +739,13 @@ gnome_entry_load_history (GnomeEntry *gentry)
 	set_combo_items (gentry);
 
 	g_slist_free (gconf_items);
-	g_object_unref (G_OBJECT (client));
 }
 
 /**
  * gnome_entry_clear_history
  * @gentry: Pointer to GnomeEntry object.
  *
- * Description:  Clears the history, you should call #gnome_entry_save_history
- * To make the change permanent.
- *
- * Returns:
+ * Description:  Clears the history.
  */
 void
 gnome_entry_clear_history (GnomeEntry *gentry)
@@ -673,15 +756,17 @@ gnome_entry_clear_history (GnomeEntry *gentry)
 	free_items (gentry);
 
 	set_combo_items (gentry);
+
+	gnome_entry_save_history (gentry);
 }
 
 static gboolean
-check_for_duplicates (GSList *gconf_items, 
+check_for_duplicates (GSList *gconf_items,
 		      const struct item *item)
 {
 
 	for (; gconf_items; gconf_items = gconf_items->next) {
-		
+
 		if (strcmp ((gchar *) gconf_items->data, item->text) == 0) {
 			return FALSE;
 		}
@@ -695,25 +780,26 @@ static void
 gnome_entry_save_history (GnomeEntry *gentry)
 {
 	GList *items;
-	GConfClient *client;
 	GSList *gconf_items;
 	struct item *item;
 	gchar *key;
 	gint n;
-	
+
 	g_return_if_fail (gentry != NULL);
 	g_return_if_fail (GNOME_IS_ENTRY (gentry));
 
-	if (gnome_program_get_app_id (gnome_program_get()) == NULL || gentry->_priv->history_id == NULL)
+	if (gnome_program_get_app_id (gnome_program_get ()) == NULL ||
+	    gentry->_priv->history_id == NULL)
 		return;
 
-	key = build_gconf_key (gentry);
-	gnomeui_gconf_lazy_init ();
-	client = gconf_client_get_default ();
+	g_return_if_fail (gentry->_priv->gconf_client != NULL);
 
+	key = build_gconf_key (gentry);
 	gconf_items = NULL;
 
-	for (n = 0, items = gentry->_priv->items; items && n < gentry->_priv->max_saved; items = items->next, n++) {
+	for (n = 0, items = gentry->_priv->items;
+	     items && n < gentry->_priv->max_saved;
+	     items = items->next, n++) {
 		item = items->data;
 
 		if (item->save && check_for_duplicates (gconf_items, item)) {
@@ -721,11 +807,13 @@ gnome_entry_save_history (GnomeEntry *gentry)
 		}
 	}
 
+	gconf_items = g_slist_reverse (gconf_items);
+
 	/* Save the list */
-	gconf_client_set_list (client, key, GCONF_VALUE_STRING, gconf_items, NULL);
-	
+	gentry->_priv->saving_history = TRUE;
+	gconf_client_set_list (gentry->_priv->gconf_client, key, GCONF_VALUE_STRING, gconf_items, NULL);
+
 	g_free (key);
-	g_object_unref (G_OBJECT (client));
 }
 
 static void
@@ -801,10 +889,66 @@ get_position (GtkEditable    *editable)
 	return gtk_editable_get_position (GTK_EDITABLE (entry));
 }
 
+/* Copied from gtkentry */
+static void
+do_insert_text (GtkEditable *editable,
+		const gchar *new_text,
+		gint         new_text_length,
+		gint        *position)
+{
+	GtkEntry *entry = GTK_ENTRY (gnome_entry_gtk_entry (GNOME_ENTRY (editable)));
+	gchar buf[64];
+	gchar *text;
+
+	if (*position < 0 || *position > entry->text_length)
+		*position = entry->text_length;
+
+	g_object_ref (G_OBJECT (editable));
+
+	if (new_text_length <= 63)
+		text = buf;
+	else
+		text = g_new (gchar, new_text_length + 1);
+
+	text[new_text_length] = '\0';
+	strncpy (text, new_text, new_text_length);
+
+	g_signal_emit_by_name (editable, "insert_text", text, new_text_length, position);
+
+	if (new_text_length > 63)
+		g_free (text);
+
+	g_object_unref (G_OBJECT (editable));
+}
+
+/* Copied from gtkentry */
+static void
+do_delete_text (GtkEditable *editable,
+		gint         start_pos,
+		gint         end_pos)
+{
+	GtkEntry *entry = GTK_ENTRY (gnome_entry_gtk_entry (GNOME_ENTRY (editable)));
+
+	if (end_pos < 0 || end_pos > entry->text_length)
+		end_pos = entry->text_length;
+	if (start_pos < 0)
+		start_pos = 0;
+	if (start_pos > end_pos)
+		start_pos = end_pos;
+
+	g_object_ref (G_OBJECT (editable));
+
+	g_signal_emit_by_name (editable, "delete_text", start_pos, end_pos);
+
+	g_object_unref (G_OBJECT (editable));
+}
+
 static void
 gnome_entry_editable_init (GtkEditableClass *iface)
 {
 	/* Just proxy to the GtkEntry */
+	iface->do_insert_text = do_insert_text;
+	iface->do_delete_text = do_delete_text;
 	iface->insert_text = insert_text;
 	iface->delete_text = delete_text;
 	iface->get_chars = get_chars;
